"""This extension generates the requestctl tool."""

from argparse import ArgumentParser, Namespace, SUPPRESS
import logging
import sys


from .cli import Requestctl, SCOPE_TO_ENTITY
from .error import RequestctlError
from .schema import SCHEMA, SYNC_ENTITIES

# public api
from .schema import get_schema  # noqa: F401


def parse_args(args) -> Namespace:
    """Parse command-line arguments."""
    parser = ArgumentParser(
        "requestctl",
        description="Tool to control/ratelimit/ban web requests dynamically",
    )
    parser.add_argument(
        "--config", "-c", help="Configuration file", default="/etc/conftool/config.yaml"
    )
    parser.add_argument("--debug", action="store_true")
    command = parser.add_subparsers(help="Command to execute", dest="command")
    command.required = True
    # Sync command
    # Synchronize entities from yaml files to the datastore
    # Example: requestctl sync --purge ipblock .
    sync = command.add_parser("sync", help="Synchronize data in the git repo to etcd.")
    sync.add_argument(
        "--git-repo",
        "-g",
        help="location on disc of the git repository",
        required=True,
        default="/var/lib/git/operations/private/requestctl",
    )
    sync.add_argument("object_type", help="What object type to sync", choices=SYNC_ENTITIES)
    sync.add_argument("--purge", "-p", help="Also delete removed objects.", action="store_true")
    sync.add_argument(
        "--interactive",
        "-i",
        help="Interactively sync objects if needed.",
        action="store_true",
    )
    # Validate command. Validates that the contents of a directory are all valid and syncable to
    # the datastore. Useful for CI purposes.
    validate = command.add_parser(
        "validate", help="Validate the contents of a directory are valid/syncable."
    )
    validate.add_argument("basedir", help="The directory to check for files.")
    # hidden argument used just for compliance with requestctl
    parser.add_argument("--object_type", default="action", help=SUPPRESS)
    # Dump command. Dumps the datastore to a directory that can be used with sync.
    dump = command.add_parser(
        "dump",
        help="Dumps the content of the datastore to a format that can be used by sync.",
    )
    dump.add_argument(
        "--git-repo", "-g", help="location on disc of the git repository", required=True
    )
    dump.add_argument("object_type", help="What object type to sync", choices=SYNC_ENTITIES)
    # Enable command. Enables a request action.
    enable = command.add_parser("enable", help="Turns on a specific action")
    enable.add_argument(
        "--scope",
        "-s",
        help="What system to search the action for",
        choices=SCOPE_TO_ENTITY.keys(),
        default="varnish",
    )
    enable.add_argument("action", help="Action to enable")
    # Disable command. Disables a request action
    disable = command.add_parser("disable", help="Turns off a specific action")
    disable.add_argument(
        "--scope",
        "-s",
        help="What object type to disable",
        choices=SCOPE_TO_ENTITY.keys(),
        default="varnish",
    )
    disable.add_argument("action", help="Action to enable")
    # Commit command. Actually compiles the enabled actions to VCL.
    commit = command.add_parser("commit", help="Actually write your changes to the edges.")
    commit.add_argument(
        "--batch",
        "-b",
        help="Does not ask for confirmation before committing",
        action="store_true",
    )
    # Get command
    # Gets either all or one specific object from the datastore, outputs in various formats
    # Examples:
    # requestctl get action
    # requestctl get action cache-text/block_cloud
    # requestctl get action cache-text/block_cloud -o yaml
    get = command.add_parser("get", help="Get an object")
    get.add_argument("object_type", help="What objects to get", choices=SCHEMA.keys())
    get.add_argument("object_path", help="The full name of the object", nargs="?", default="")
    get.add_argument(
        "-o",
        "--output",
        help="Choose the format for output: pretty, json, yaml. "
        "Pretty output is disabled for actions at the moment.",
        choices=["pretty", "json", "yaml"],
        default="pretty",
    )
    # Log command. Outputs a typical varnishncsa command to log the selected action
    log = command.add_parser("log", help="Get the varnishncsa to log requests matching an object.")
    log.add_argument(
        "object_path",
        help="The full name of the object",
    )
    # vcl command. Outputs the VCL for the selected action.
    vcl = command.add_parser(
        "vcl", help="Get the varnish VCL that will be generated by this action."
    )
    vcl.add_argument(
        "object_path",
        help="The full name of the object",
    )
    # haproxycfg command. Outputs the haproxy configuration for the selected action.
    haproxycfg = command.add_parser(
        "haproxycfg",
        help="Get the haproxy configuration that will be generated by this haproxy_action.",
    )
    haproxycfg.add_argument(
        "object_path",
        help="The full name of the object",
    )
    # find command. Returns the actions that include a specific pattern/ipblock
    find = command.add_parser("find", help="Find which actions include a specific pattern/ipblock")
    # Scope is none by default, as we might want to search both varnish and haproxy actions.
    find.add_argument(
        "--scope",
        "-s",
        help="What system to search the action for",
        choices=SCOPE_TO_ENTITY.keys(),
    )
    find.add_argument(
        "search_string",
        help="The string to search in the expression. Must be in the format <scope>/<name>."
        "No regex matching or partial string match is performed.",
    )
    # find-ip command. Returns all the ipblocks and IP belongs to, if any
    find_ip = command.add_parser(
        "find-ip",
        help="Find if an IP is part of any CIDR of any ipblock definitions on disk.",
    )
    find_ip.add_argument(
        "--git-repo",
        "-g",
        help="location on disc of the git repository",
        default="/srv/private/requestctl",
    )
    find_ip.add_argument(
        "ip",
        help="The IP address to search for.",
    )
    return parser.parse_args(args)


def main():
    """Run the tool."""
    logger = logging.getLogger("reqctl")
    options = parse_args(sys.argv[1:])
    rq = Requestctl(options)
    try:
        # TODO: add support to let a custom exit code surface, for example for
        # "failed successfully" operations
        rq.run()
    except RequestctlError as e:
        logger.error(e)
        sys.exit(1)
